import json
import os

from os.path import abspath, join

"""Utility to update files when data changes

MAYBE make oop instead of static
MAYBE use multiprocessing

Methods
-------
update_json
    update json file with data from dictionary
write_file
    write file
delete_file
    delete file
delete_command_file
    delete file containing command
get_file_by_command
    get file containing command
"""


def update_json(dictionary, file_path):
    """update json file with data from dictionary

    Parameters
    ----------
    dictionary : dict
        dict containing updated data
    file_path : str
        path of file to be updated
    """

    with open(file_path, 'w') as file:
        json.dump(dictionary, file, indent=2)


def write_file(content, file_path):
    """write file

    Parameters
    ----------
    content : str
        content of file
    file_path : str
        relative file path
    """
    with open(file_path, 'w') as file:
        file.write(content)


def delete_file(file):
    """delete file

    Parameters
    ----------
    file : str
        file to be deleted
    """

    os.remove(file)


def delete_command_file(command):
    """delete file containing command

    Parameters
    ----------
    command : Command
        command contained by file to be deleted
    """

    delete_file(get_file_by_command(command))


def get_file_by_command(command):
    """get file containing command

    Parameters
    ----------
    command : Command
        command contained by file
    """

    return join(*abspath(command.__module__).split('.')) + '.py'



import logging

from symbot.chat.message import Message
from symbot.control.control import Control
from symbot.dev.commands._base_command import BaseCommand


class Command(BaseCommand):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = '!howbad'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):
        try:
            broadcaster = self.control.environment.get('broadcaster')
        except KeyError:
            logging.info(f'{self.name} unable to find var broadcaster')
            return
        try:
            bad = self.control.environment.get('bad')
        except KeyError:
            logging.info(f'{self.name} unable to find var bad')
            return

        response = f'{broadcaster} has done {bad} things poorly'
        await self.control.respond(response)



import logging

from symbot.chat.message import Message
from symbot.control.control import Control
from symbot.dev.commands._base_command import BaseCommand


class Command(BaseCommand):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = '!uptime'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        uptime = self.control.get_media('uptime')

        if uptime:

            elapsed = await uptime.run()

            seconds = elapsed % 60
            elapsed = (elapsed - seconds) / 60
            minutes = int(elapsed) % 60
            elapsed = (elapsed - minutes) / 60
            hours = int(elapsed) % 24
            elapsed = (elapsed - hours) / 24
            days = int(elapsed)

            time_str = ''
            if seconds > 0:
                time_str = f' {seconds} s'
            if minutes > 0:
                time_str = f' {minutes} m' + time_str
            if hours > 0:
                time_str = f' {hours} h' + time_str
            if days > 0:
                time_str = f' {days} d' + time_str

            response = f'Stream has been live for{time_str}'

            await self.control.respond(response)

        else:

            logging.info('Required media unavailable: uptime')



from symbot.control.control import Control
from symbot.dev.media._base_media import BaseMedia

from time import time


class Media(BaseMedia):
    """Trace time since chat bot was started"""

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = 'uptime'
        self.author = 'fd_symbicort'
        self.start = time()

    async def run(self, **kwargs):
        return int(time() - self.start)



from symbot.chat.message import Message
from symbot.control.control import Control
from symbot.dev.commands._base_command import BaseCommand


class Command(BaseCommand):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = '!well'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        msg.command = '!good'

        await self.control.requeue(msg)



from os.path import join, dirname, abspath


# directories
data_path = join(dirname(abspath(__file__)), 'data')
dev_path = join(dirname(abspath(__file__)), 'dev')

# server connection
server = 'irc.chat.twitch.tv'
port = 6667

# connection credentials
# get your token on https://twitchapps.com/tmi/
channel = 'broadcaster username'
nick = 'bot username'
token = 'oauth:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'



import logging

from symbot.chat.message import Message
from symbot.dev.meta._base_meta_command import BaseMetaCommand


class Command(BaseMetaCommand):
    """Manipulate variables in the environment

    Methods
    -------
    run
        execute when !var is called
    getvar
        retrieve variable from environment
    setvar
        edit a command
    delvar
        delete a command
    """

    def __init__(self, control):
        super().__init__(control)
        self.name = '!var'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        try:
            operation = msg.context[0]
        except IndexError:
            logging.info(f'{self.name} missing operation argument')
            return

        if operation == 'get':
            await self.getvar(msg)
        elif operation == 'set':
            await self.setvar(msg)
        elif operation == 'del':
            await self.delvar(msg)
        else:
            logging.info(f'{self.name} encountered an undefined operation argument {operation}')

    async def getvar(self, msg):
        """retrieve variable from environment

        Parameters
        ----------
        msg : Message
            user message trying to retrieve a variable
        """

        try:
            var = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} missing context var')
            return
        try:
            val = self.control.environment.get(var)
        except KeyError:
            logging.info(f'{self.name} unable to find var {var}')
            return

        response = f'{var} has value {val}'
        await self.control.respond(response)

    async def setvar(self, msg):
        """set value of a variable

        Parameters
        ----------
        msg : Message
            user message trying to set the value of a variable
        """

        try:
            var = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} missing context var')
            return
        try:
            new_value = msg.context[2]
        except IndexError:
            logging.info(f'{self.name} missing context value')
            return
        self.control.environment.set(var, new_value)

        response = f'value of {var} has been changed to {new_value}'
        await self.control.respond(response)

    async def delvar(self, msg):
        """delete a variable from the environment

        Parameters
        ----------
        msg : Message
            message containing var to be deleted
        """

        try:
            var = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} missing context var')
            return
        try:
            self.control.environment.delete(var)
        except KeyError:
            logging.info(f'{self.name} unable to find var {var}')
            return

        response = f'{var} has been deleted'
        await self.control.respond(response)



import re
import time


class Message:
    """Represents a message received from Twitch as workable object

    Attributes
    ----------
    user : str
        user who sent the message
    channel : str
        channel message was sent to
    timestamp: float
        time since epoch
    content : str
        actual message sent by user
    command : str
        command identifier
    context : str
        context to a command
    """

    def __init__(self, received):
        """
        Parameters
        ----------
         received : str
            private message from Twitch channel
        """

        # regex parser that works on private messages from twitch channel
        # MAYBE try/catch for correct input just in case
        # MAYBE expand to generic messages
        groups = re.search(
            ':(.*)!.*@.*\.tmi\.twitch\.tv PRIVMSG (#.*)? :(.*)',
            received).groups()

        self.user = groups[0]
        self.channel = groups[1]
        self.timestamp = time.time()
        self.content = groups[2]
        # extract command and context from content
        split = self.content.split(' ')
        self.command = split[0]
        self.context = split[1:]



import asyncio
import logging
import sys

from symbot.chat.chat import Chat
from symbot.control.control import Control


async def main():
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s: %(message)s',
        handlers=[logging.StreamHandler(sys.stdout)]
    )

    chat = Chat()
    control = Control()
    control.msg_queue = chat.msg_queue
    chat.resp_queue = control.resp_queue

    await chat.open_connection()
    await asyncio.gather(chat.read(), control.process(), chat.write())


if __name__ == '__main__':
    asyncio.run(main())



import logging

from symbot.chat.message import Message
from symbot.dev.meta._base_meta_command import BaseMetaCommand


class Command(BaseMetaCommand):

    def __init__(self, control):
        super().__init__(control)
        self.name = '!setuserperm'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        try:
            user = msg.context[0]
        except IndexError:
            logging.info(f'{self.name} missing context user')
            return
        try:
            new_value = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} missing context value')
            return
        try:
            if not int(new_value) in [0, 1, 2, 3, 4]:
                logging.info(f'{self.name} requires value in [0, 1, 2, 3, 4]')
                return
        except ValueError:
            logging.info(f'{self.name} requires value in [0, 1, 2, 3, 4]')
            return
        self.control.permissions.set(user, int(new_value))

        response = f'{user} has now permission level {new_value}'
        await self.control.respond(response)



from symbot.chat.message import Message
from symbot.control.control import Control
from symbot.dev.commands._base_command import BaseCommand


class Command(BaseCommand):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = 'hey'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        recipient = msg.user

        response = f'Hello there, {recipient} HeyGuys'
        await self.control.respond(response)



import json
import logging
from os.path import join

from symbot.config import data_path
from symbot.util.updater import update_json


class Permissions:
    """Controller class for user permission levels

    user permission levels as of now:
        0 : broadcaster
        1 : moderator
        2 : whitelisted
        3 : default
        4 : blacklisted

    Attributes
    ----------
    file_path : str
        path of datafile
    permissions : dict
        dict to store user permission levels

    Methods
    -------
    set
        assign permission level to a user
    get
        retrieve user permission level
    check_call
        determine whether user has sufficient permission to call a command
    check_meta
        determine whether user has sufficient permission to modify a command
    """

    def __init__(self):

        # MAYBE put path into config or somewhere else
        self.file_path = join(data_path, 'user_permissions.json')

        # load permissions from data folder
        try:
            logging.info('loading user permissions')
            with open(self.file_path) as file:
                self.permissions = json.load(file)
        # or start a fresh environment
        except FileNotFoundError:
            logging.info(f'no permissions found in {self.file_path}')
            self.permissions = {}
            logging.info('created new user permissions')

    def set(self, user, level):
        """assign permission level to a user

        Parameters
        ----------
        user : str
            user identifier
        level : int
            permission level
        """

        if level == 3:
            del self.permissions[user]
        else:
            self.permissions[user] = level
        update_json(self.permissions, self.file_path)

    def get(self, user):
        """retrieve user permission level

        Parameters
        ----------
        user : str
            user identifier

        Returns
        -------
        int
            user permission level
        """

        if user in self.permissions:
            # get user level from data
            return self.permissions[user]
        else:
            # or use default permission level
            # MAYBE enable dev user levels from config
            # MAYBE use enumerate?
            return 3

    def check_call(self, cmd_level, user):
        """determine whether user has sufficient permission to call a command

        Parameters
        ----------
        cmd_level : int
            minimum permission level required by command
        user : str
            user name

        Returns
        -------
        bool
            user has sufficient permission
        """

        return self.get(user) <= cmd_level

    def check_meta(self, cmd_level, user):
        """determine whether user has sufficient permission to modify a command

        Parameters
        ----------
        cmd_level : int
            minimum permission level required by command
        user : str
            user name

        Returns
        -------
        bool
            user has sufficient permission
        """

        return self.get(user) < cmd_level



import re
from os.path import join

from symbot.util import updater
from symbot.util.strings import stringify
from symbot.config import dev_path


class Builder:
    """ Builder to assemble or reassemble commands, write files and load them

    Methods
    -------
    create_command
        creates command from a skeleton as file and loads it
    edit_command
        recreates command from a skeleton as file and reloads it
    delete_command
        deletes command and unloads it
    assemble
        assembles command code from skeleton
    """

    def __init__(self, control):
        self.control = control

    def create_command(self, skeleton):
        """creates command from a skeleton as file and loads it

        Parameters
        ----------
        skeleton : dict
            blueprint containing all information about the command
        """

        # assemble command
        code = self.assemble(skeleton)
        file_name = re.search(r'\w+', skeleton['settings']['name']).group(0)
        file_path = join(dev_path, 'commands', 'new', file_name + '.py')
        # write file
        updater.write_file(code, file_path)
        # load command
        self.control.import_command(file_path)

    def edit_command(self, command, skeleton, path):
        """recreates command from a skeleton as file and reloads it

        Parameters
        ----------
        command : Command
            command to be edited
        skeleton : dict
            blueprint containing all information about the command
        path : str
            file path of module containing command
        """

        # assemble command
        code = self.assemble(skeleton)
        # rewrite file
        updater.write_file(code, path)
        # reload command
        self.control.delete_command(command)
        self.control.import_command(path)

    def delete_command(self, command):
        """deletes command and unloads it

        Parameters
        ----------
        command : Command
            command to be deleted
        """

        # delete file
        updater.delete_command_file(command)
        # unload command
        self.control.delete_command(command)

    def assemble(self, skeleton):
        """assembles command code from skeleton

        Parameters
        ----------
        skeleton : dict
            blueprint containing all information about the command
        """

        # header always the same
        code = \
            'import logging\n' \
            '\n' \
            'from symbot.chat.message import Message\n' \
            'from symbot.control.control import Control\n' \
            'from symbot.dev.commands._base_command import BaseCommand\n' \
            '\n' \
            '\n' \
            'class Command(BaseCommand):\n' \
            '\n' \
            '    def __init__(self, control: Control):\n' \
            '        super().__init__(control)\n'

        # override default settings
        for setting, value in skeleton['settings'].items():
            code += \
                f'        self.{setting} = {value}\n'

        code += \
            '\n' \
            '    async def run(self, msg: Message):\n'

        # adjust message according to alias
        for alias in skeleton['alias']:
            code += \
                '\n' \
                f'        msg.command = {stringify(alias)}\n' \
                '\n' \
                '        await self.control.requeue(msg)\n'

        # aliases are special case
        # ignore everything else and
        # only execute aliases
        # MAYBE change in the future
        if len(skeleton['alias']) > 0:
            return code

        # initialize var if it does not exist
        # increment vars
        for var in skeleton['c']:
            self.control.environment.initialize(var)
            code += \
                '        try:\n' \
                f'            {var} = self.control.environment.increment({stringify(var)})\n' \
                '        except KeyError:\n' \
                f'            logging.info(f\'{{self.name}} unable to find var {var}\')\n' \
                '            return\n'

        # initialize var if it does not exist
        # retrieve vars
        for var in skeleton['v']:
            self.control.environment.initialize(var)
            code += \
                '        try:\n' \
                f'            {var} = self.control.environment.get({stringify(var)})\n' \
                '        except KeyError:\n' \
                f'            logging.info(f\'{{self.name}} unable to find var {var}\')\n' \
                '            return\n'

        # retrieve arguments from message
        for i, arg in enumerate(skeleton['a']):
            code += \
                '        try:\n' \
                f'            {arg} = msg.context[{i}]\n' \
                '        except IndexError:\n' \
                f'            logging.info(f\'{{self.name}} missing context {arg}\')\n' \
                '            return\n'

        # retrieve user from message
        for user in skeleton['u']:
            code += \
                f'        {user} = msg.user\n'

        # generate response
        response = ' '.join(skeleton['r'])
        code += \
            '\n' \
            f'        response = f\'{response}\'\n'

        # respond to control
        code += \
            '        await self.control.respond(response)\n'

        return code



import logging

from symbot.chat.message import Message
from symbot.dev.meta._base_meta_command import BaseMetaCommand


class Command(BaseMetaCommand):

    def __init__(self, control):
        super().__init__(control)
        self.name = '!setcmdsetting'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        try:
            name = msg.context[0]
        except IndexError:
            logging.info(f'{self.name} missing context command name')
            return
        try:
            setting = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} missing context setting')
            return
        try:
            value = msg.context[2]
        except IndexError:
            logging.info(f'{self.name} missing context value')
            return

        if self.control.settings.set(name, setting, value):
            response = f'Changed {setting} of {name} to value {value}'
            await self.control.respond(response)



import asyncio
import logging
import os
import sys
from importlib import import_module
from os.path import join

from symbot.config import dev_path
from symbot.control.auxiliary.cooldowns import Cooldowns
from symbot.control.auxiliary.environment import Environment
from symbot.control.auxiliary.permissions import Permissions
from symbot.control.auxiliary.settings import Settings


class Control:
    """Central control element

    Controller communicates between Twitch chat, various data files,
    media elements and command calls & responses. Dynamically loads
    commands from dev and utilizes auxiliary controllers.

    Attributes
    ----------
    permissions : Permissions
        checks user permission levels
    environment : Environment
        manages environmental variables
    cooldowns : Cooldowns
        tracks command cooldowns
    settings : Settings
        overrides default command settings
    media : dict
        dict of all media loaded dynamically
    commands : list
        dict of all commands loaded dynamically
    msg_queue : Queue
        thread safe queue to receive Twitch messages
    resp_queue : Queue
        thread safe queue to push responses to

    Methods
    -------
    import_media
        recursively import and instantiate all media in a directory
    get_media
        try to find media by name
    import_commands
        recursively import and instantiate all commands in a directory
    delete_command
        delete command from dict and unload module
    get_command
        try to find command by name
    requeue
        push message back to message queue
    respond
        push response to response queue
    process
        continuously process messages from Twitch channel
    """

    def __init__(self):

        # dynamically load in media
        self.media = {}
        logging.info('loading media')
        self.import_media(join(dev_path, 'media'))

        # dynamically load in commands
        self.commands = {}
        logging.info('loading commands')
        self.import_command(join(dev_path, 'commands'))
        self.import_command(join(dev_path, 'meta'))

        # auxiliary controllers
        self.permissions = Permissions()
        self.environment = Environment()
        self.cooldowns = Cooldowns()
        self.settings = Settings(self.commands)

        # async data structures
        self.msg_queue = None
        self.resp_queue = asyncio.Queue()

    def import_media(self, path):
        """recursively import and instantiate all media in a directory

        Parameters
        ----------
        path : str
            current directory of modules to be imported
        """

        splits = path.split(os.sep)
        file = splits[-1]
        # exclude files not meant to be imported as command
        if file.startswith('_'):
            return
        # import .py module
        elif file.endswith('.py'):
            module = file[:-3]
            for m in splits[::-1][1:]:
                module = '.'.join([m, module])
                if m == 'symbot':
                    break
            media = import_module(module).Media(self)
            self.media[media.name] = media
        # import package
        else:
            for file in os.listdir(path):
                # import modules from lower levels recursively
                self.import_media(join(path, file))

    def get_media(self, media_name):
        """try to find media by name

        Parameters
        ----------
        media_name : str
            media identifier

        Returns
        -------
        Media
            desired media, or None if not found
        """

        if media_name not in self.media:
            return None
        return self.media[media_name]

    def import_command(self, path):
        """recursively import and instantiate all commands in a directory

        Parameters
        ----------
        path : str
            current directory of modules to be imported
        """

        splits = path.split(os.sep)
        file = splits[-1]
        # exclude files not meant to be imported as command
        if file.startswith('_'):
            return
        # import .py module
        elif file.endswith('.py'):
            module = file[:-3]
            for m in splits[::-1][1:]:
                module = '.'.join([m, module])
                if m == 'symbot':
                    break
            command = import_module(module).Command(self)
            self.commands[command.name] = command
        # import package
        else:
            for file in os.listdir(path):
                # import modules from lower levels recursively
                self.import_command(join(path, file))

    def delete_command(self, command):
        """delete command from dict and unload module

        Parameters
        ----------
        command : Command
            command to be deleted
        """

        del self.commands[command.name]
        sys.modules.pop(command.__module__)

    def get_command(self, cmd_name):
        """try to find command by name

        Parameters
        ----------
        cmd_name : str
            command identifier

        Returns
        -------
        Command
            desired command, or None if not found
        """

        if cmd_name not in self.commands:
            return None
        return self.commands[cmd_name]

    async def requeue(self, msg):
        """push message back to message queue

        Parameters
        ----------
        msg : Message
            modified message that needs to be reprocessed
        """
        await self.msg_queue.put(msg)

    async def respond(self, response):
        """push response to response queue

        Parameters
        ----------
        response : str
            response to be sent to Twitch channel
        """

        await self.resp_queue.put(response)

    async def process(self):
        """continuously process messages from Twitch channel

        continuously process messages from Twitch channel. Check for
        various flags before executing a command. Not every command
        generates a response.
        """

        # run forever
        while True:
            # wait for message to process
            # provided by chat
            msg = await self.msg_queue.get()
            command = self.get_command(msg.command)
            # check for existence
            if not command:
                continue
            # check if command is enabled
            if not command.enabled:
                logging.info(f'{command.name} is disabled')
                continue
            # check for permission
            if not self.permissions.check_call(command.permission_level, msg.user):
                logging.info(f'{msg.user} has insufficient permission to call {command.name}')
                continue
            # check for cooldown
            if self.cooldowns.has_cooldown(command, msg.timestamp):
                logging.info(f'{command.name} is still on cooldown')
                continue
            # command is safe to execute
            # append command to asyncio loop
            asyncio.get_running_loop().create_task(command.run(msg))



import json
import logging
from os.path import join

from symbot.config import data_path
from symbot.util.updater import update_json


class Environment:
    """Controller class for Twitch chat's environmental variables

    Attributes
    ----------
    environment : dict
        dict to store environmental variables

    Methods
    -------
    get
        get the value of a variable
    set
        set the value of a variable
    initialize
        initialize new variable as 0 if it doesn't exist yet
    increment
        increment the value of a variable
    delete
        delete variable from environment
    """

    def __init__(self):

        # MAYBE put path into config or somewhere else
        self.file_path = join(data_path, 'environment.json')

        # load environment from data folder
        try:
            logging.info('loading environment')
            with open(self.file_path) as file:
                self.environment = json.load(file)
        # or start a fresh environment
        except FileNotFoundError:
            logging.info(f'no environment found in {self.file_path}')
            self.environment = {}
            logging.info('created new environment')

    def get(self, var):
        """get the value of a variable

        Parameters
        ----------
        var : str
            variable name identifying desired value

        Returns
        -------
        object
            desired value
        """

        # handle var not defined in command
        return self.environment[var]

    def set(self, var, val):
        """set the value of a variable

        Parameters
        ----------
        var : str
            variable name identifying value
        val : object
            value to be set

        Returns
        -------
        object
            set value
        """

        # if var already exists, typecast new value to old type
        if var in self.environment:
            val = type(self.environment[var])(val)
        # if var is a digit, typecast to int
        elif type(val) == str:
            if val.isdigit():
                val = int(val)

        self.environment[var] = val
        update_json(self.environment, self.file_path)
        return val

    def initialize(self, var):
        """initialize new variable as 0 if it doesn't exist yet

        Parameters
        ----------
        var : str
            variable to be initiated
        """

        if var not in self.environment:
            self.set(var, 0)

    def increment(self, var):
        """increment the value of a variable

        Parameters
        ----------
        var : str
            variable name identifying value to be incremented

        Returns
        -------
        int
            incremented value
        """

        # handle var not defined in command
        self.environment[var] += 1
        update_json(self.environment, self.file_path)
        return self.environment[var]

    def delete(self, var):
        """delete variable from environment

        Parameters
        ----------
        var : str
            variable to be deleted from environment
        """

        del self.environment[var]



from abc import ABC, abstractmethod

from symbot.control.control import Control


class BaseMedia(ABC):
    """Abstract base class to implement new commands

    Abstract base class developers need to implement for new commands.
    Enforces all properties and methods required by controller.

    Attributes
    ----------
    control : Control
        central control element
    name : str
        media identifier (default: module name)
    author : str
        author identifier (default: None)
    permission_level : int
        permission level of media (default: public)
    cooldown : float
        minimum seconds between command calls (default: 0)
    enabled : bool
        command is enabled

    Methods
    -------
    run
        execute media under consideration of arguments
    """

    def __init__(self, control: Control):
        self.control = control
        self.name = self.__module__.split('.')[-1]
        self.author = ''
        self.permission_level = 3
        self.cooldown = 0
        self.enabled = True

    @abstractmethod
    async def run(self, **kwargs):
        """execute media under consideration of arguments

        Parameters
        ----------
        kwargs : dict
            key-worded arguments
        """
        ...



import logging

from symbot.chat.message import Message
from symbot.dev.meta._base_meta_command import BaseMetaCommand
from symbot.util import updater


class Command(BaseMetaCommand):
    """Purge all commands created by a specified user"""

    def __init__(self, control):
        super().__init__(control)
        self.name = '!purge'
        self.author = 'fd_symbicort'
        self.permission_level = 0

    async def run(self, msg: Message):

        try:
            user = msg.context[0]
        except IndexError:
            logging.info(f'{self.name} missing context user name')
            return

        # gather commands created by user
        commands = [command for command in self.control.commands.values() if command.author == user]

        # if any such commands exist
        if commands:

            for command in commands:
                # delete command from control
                self.control.delete_command(command)
                # delete file containing command
                updater.delete_command_file(command)

            command_names = ', '.join([command.name for command in commands])

            response = f'Purged commands: {command_names} of user {user}'
        else:
            response = f'{user} has not created any commands'

        await self.control.respond(response)



import json
import logging
from os.path import join

from symbot.config import data_path
from symbot.util.updater import update_json


class Settings:
    """Controller class to override default command settings

    Attributes
    ----------
    commands : dict
        commands loaded by central control
    file_path : str
        path of datafile
    settings : dict
        dict to store command settings

    Methods
    -------
    override
        override default settings from devs with user settings
    set_attr
        set an attribute of a command
    """

    def __init__(self, commands):
        self.commands = commands

        # MAYBE put path into config or somewhere else
        self.file_path = join(data_path, 'command_settings.json')

        # load command settings from data folder
        try:
            logging.info('loading command settings')
            with open(self.file_path) as file:
                self.settings = json.load(file)
        # or start fresh command settings
        except FileNotFoundError:
            logging.info(f'no command settings found in {self.file_path}')
            self.settings = {}
            logging.info('created new command settings')

        # override default settings with user settings
        self.override()

    def override(self):
        """override default settings from devs with user settings"""

        logging.info('overriding default command settings with user settings')
        # for each cmd_name, look up its settings
        for cmd_name, setting in self.settings.items():
            # and override the command attr for each altered attr
            for attr, val in setting.items():
                self.set_attr(cmd_name, attr, val)

    def set_attr(self, cmd_name, attr, val):
        """set an attribute of a command

        Parameters
        ---------
        cmd_name : str
            command identifier
        attr : str
            attribute to be set
        val : object
            value to be assigned

        Returns
        -------
        bool
            assignment was successful
        """

        try:
            command = self.commands[cmd_name]
        except KeyError:
            logging.info(f'can not find command {cmd_name}')
            return False

        try:
            cmd_attr = command.__getattribute__(attr)
            # check for str to bool casting shenanigans
            if type(cmd_attr) == bool and type(val) == str:
                val = val.lower() == 'true'
            # enforce that val has to be same type as cmd_attr
            val = (type(cmd_attr))(val)
            command.__setattr__(attr, val)
            return True
        except AttributeError:
            logging.info(f'{cmd_name} unable to set '
                         f'{attr}, because '
                         f'command has no such attribute')
            return False
        except ValueError:
            logging.info(f'{cmd_name} unable to set '
                         f'{attr}, because '
                         f'{val} is of type '
                         f'{type(val)} instead of type'
                         f'{type(cmd_attr)}')
            return False

    def set(self, cmd_name, attr, val):
        """set an attribute of a command and update settings data

        Parameters
        ---------
        cmd_name : str
            command identifier
        attr : str
            attribute to be set
        val : object
            value to be assigned
        """

        if self.set_attr(cmd_name, attr, val):
            if cmd_name in self.settings:
                self.settings[cmd_name][attr] = val
            else:
                self.settings[cmd_name] = {attr: val}
            update_json(self.settings, self.file_path)
            return True
        return False



import logging

from symbot.chat.message import Message
from symbot.control.control import Control
from symbot.dev.commands._base_command import BaseCommand


class Command(BaseCommand):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = '!highfive'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        try:
            recipient = msg.context[0]
        except IndexError:
            logging.info(f'{self.name} missing context recipient')
            return
        user = msg.user

        response = f'{user} gave {recipient} a highfive!'
        await self.control.respond(response)



from symbot.control.control import Control
from symbot.dev.media._base_media import BaseMedia

import time


class Media(BaseMedia):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = 'localtime'
        self.author = 'fd_symbicort'

    async def run(self, **kwargs):
        return time.strftime('%I:%M %p', time.localtime(time.time()))



"""Utility to manipulate strings

Methods
-------
stringify
     wraps string in apostrophes

"""


def stringify(s):
    """wraps string in apostrophes

    Parameters
    ----------
    s : str
        string to be wrapped in apostrophes

    Returns
    -------
    str
        string wrapped in apostrophes
    """
    return '\'' + s + '\''



import logging
import re

from symbot.chat.message import Message
from symbot.dev.meta._base_meta_command import BaseMetaCommand
from symbot.dev.meta._builder import Builder
from symbot.util import updater
from symbot.util.strings import stringify


class Command(BaseMetaCommand):
    """Central meta command

    Add, Edit or Delete commands and their files with !command

    Methods
    -------
    run
        execute when !command is called
    addcom
        add a new command
    editcom
        edit a command
    delcom
        delete a command
    skellify_message
        parse message to blueprint of a command
    skellify_command
        parse command to blueprint of a command
    """

    def __init__(self, control):
        super().__init__(control)
        self.name = '!command'
        self.author = 'fd_symbicort'

        self.builder = Builder(control)

        # filter string escapes
        self.msg_filter = re.compile('[\'\"\\\]')

        # message extractors
        self.arg_extractor = re.compile('\\$(.*){(.*)}')
        self.setting_extractor = re.compile('-(.*)=(.*)')

        # file extractors
        self.r_extractor = re.compile(r'response = f\'(.*)\'')
        self.v_extractor = re.compile(r'get\(\'(\w*)\'\)')
        self.c_extractor = re.compile(r'increment\(\'(\w*)\'\)')
        self.a_extractor = re.compile(r' (\w*) = msg\.context\[\d\]')
        self.u_extractor = re.compile(r' (\w*) = msg\.user')
        self.alias_extractor = re.compile(r'msg\.command = \'(\S*)\'')
        self.name_extractor = re.compile(r'name = (\S*)')
        self.author_extractor = re.compile(r'author = (\S*)')
        self.permission_level_extractor = re.compile(r'permission_level = (\S*)')
        self.cooldown_extractor = re.compile(r'cooldown = (\S*)')
        self.enabled_extractor = re.compile(r'enabled = (\S*)')

    async def run(self, msg: Message):

        try:
            operation = msg.context[0]
        except IndexError:
            logging.info(f'{self.name} missing operation argument')
            return

        if operation == 'add':
            await self.addcom(msg)
        elif operation == 'edit':
            await self.editcom(msg)
        elif operation == 'del':
            await self.delcom(msg)
        else:
            logging.info(f'{self.name} encountered an undefined operation argument {operation}')

    async def addcom(self, msg):
        """add a new command

        Parameters
        ----------
        msg : Message
            user message trying to add a new command
        """

        # extract command name
        try:
            name = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} add missing command name argument')
            return
        if self.control.get_command(name):
            logging.info(f'{self.name} add {name} already exists')
            return
        try:
            msg.context[2]
        except IndexError:
            logging.info(f'{self.name} add {name} missing content')
            return

        # create command blueprint
        skeleton = self.skellify_message(msg, name, 'add')
        if skeleton:
            # create command, load it and respond
            self.builder.create_command(skeleton)
            await self.control.respond(f'{msg.user} has added {name} to commands')

    async def editcom(self, msg):
        """edit a command

        Parameters
        ----------
        msg : Message
            user message trying to edit a command
        """

        # extract command name
        try:
            name = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} edit missing command name argument')
            return
        if not self.control.get_command(name):
            logging.info(f'{self.name} edit {name} does not exist')
            return
        try:
            msg.context[2]
        except IndexError:
            logging.info(f'{self.name} edit {name} missing content')
            return

        # skellify message and command
        skeleton_message = self.skellify_message(msg, name, 'edit')
        file = updater.get_file_by_command(self.control.get_command(msg.context[1]))
        skeleton_command = self.skellify_command(file)

        # check if only settings are being edited
        # do not change response in that case
        only_settings = True
        for k, v in skeleton_message.items():
            if k != 'settings' and v:
                only_settings = False
                break

        # overwrite settings
        del skeleton_message['settings']['name']
        del skeleton_message['settings']['author']
        for setting, value in skeleton_message['settings'].items():
            skeleton_command['settings'][setting] = value
        del skeleton_message['settings']

        # overwrite response
        if not only_settings:
            for k, v in skeleton_message.items():
                skeleton_command[k] = v

        # edit command, reload it and respond
        self.builder.edit_command(self.control.get_command(name), skeleton_command, file)
        await self.control.respond(f'{msg.user} has modified command {name}')

    async def delcom(self, msg):
        """delete a command

        Parameters
        ----------
        msg : Message
            message containing command to be deleted
        """

        # extract command name
        try:
            name = msg.context[1]
        except IndexError:
            logging.info(f'{self.name} del missing command name argument')
            return
        command = self.control.get_command(name)
        if not command:
            logging.info(f'{self.name} del {name} does not exists')
            return

        # check for safety conditions
        # MAYBE increase security
        if self.control.permissions.check_meta(command.permission_level, msg.user) \
                or command.author == msg.user:
            # delete command and respond
            self.builder.delete_command(command)
            await self.control.respond(f'{msg.user} has removed command {name}')

    def skellify_message(self, msg, name, operation):
        """parse message to blueprint of a command

        Parameters
        ----------
        msg : Message
            message to be parsed
        name : str
            command name
        operation : str
            command operation for logging

        Returns
        -------
        dict
            blueprint of a command
        """

        # blueprint container
        skeleton = {
            # response
            'r': [],
            # variables
            'v': [],
            # counters
            'c': [],
            # arguments
            'a': [],
            # user
            'u': [],
            # alias
            'alias': [],
            # settings
            'settings': {'name': stringify(name), 'author': stringify(msg.user)}
        }

        # start parsing after command name
        for s in msg.context[2:]:

            # filter string escapes [' " \] for security
            s = self.msg_filter.sub('', s)

            # $ indicates special item
            if s.startswith('$'):
                try:
                    # extract special item
                    arg, value = self.arg_extractor.search(s).groups()
                    skeleton[arg].append(value)
                    skeleton['r'].append('{' + value + '}')
                except AttributeError:
                    logging.info(f'{self.name} {operation} {name} has bad argument')
                    return
                except KeyError:
                    logging.info(f'{self.name} {operation} {name} encountered undefined argument')
                    return
            # - indicates setting
            elif s.startswith('-'):
                try:
                    # extract setting
                    setting, value = self.setting_extractor.search(s).groups()
                    if setting == 'ul':
                        skeleton['settings']['permission_level'] = int(value)
                    elif setting == 'cd':
                        skeleton['settings']['cooldown'] = float(value)
                    elif setting == 'on':
                        skeleton['settings']['enabled'] = value.lower() == 'true'
                    else:
                        logging.info(f'{self.name} {operation} {name} invalid setting {setting}')
                        return
                except AttributeError:
                    logging.info(f'{self.name} {operation} {name} has bad setting')
                    return
                except ValueError:
                    logging.info(f'{self.name} {operation} {name} can not convert setting value')
                    return
            else:
                skeleton['r'].append(s)

        # if parsing was successful, return blueprint
        # otherwise null was already returned
        return skeleton

    def skellify_command(self, path):
        """parse command to blueprint of a command

        Parameters
        ----------
        path : str
            absolute path of module containing command

        Returns
        -------
        dict
            blueprint of a command
        """

        with open(path, 'r') as f:
            code = f.read()

        skeleton = {}

        response = self.r_extractor.findall(code)
        if response:
            skeleton['r'] = response[0].split(' ')
        else:
            skeleton['r'] = []

        # variables
        skeleton['v'] = self.v_extractor.findall(code)
        # counters
        skeleton['c'] = self.c_extractor.findall(code)
        # arguments
        skeleton['a'] = self.a_extractor.findall(code)
        # user
        skeleton['u'] = self.u_extractor.findall(code)
        # alias
        skeleton['alias'] = self.alias_extractor.findall(code)

        skeleton['settings'] = {}

        name = self.name_extractor.findall(code)
        if name:
            skeleton['settings']['name'] = name[0]
        author = self.author_extractor.findall(code)
        if author:
            skeleton['settings']['author'] = author[0]
        permission_level = self.permission_level_extractor.findall(code)
        if permission_level:
            skeleton['settings']['permission_level'] = permission_level
        cooldown = self.cooldown_extractor.findall(code)
        if cooldown:
            skeleton['settings']['cooldown'] = cooldown
        enabled = self.enabled_extractor.findall(code)
        if enabled:
            skeleton['settings']['enabled'] = enabled

        return skeleton



from time import time


class Cooldowns:
    """Controller class to keep track of command cooldowns

    Attributes
    ----------
    cooldowns : dict([Command, float])
        dict to keep track of command cooldowns

    Methods
    -------
    has_cooldown
        determine whether a command is on cooldown
    """

    def __init__(self):
        self.cooldowns = dict()

    def has_cooldown(self, command, msg_timestamp):
        """determine whether a command is on cooldown

        Parameters
        ----------
        command : Command
            command to be checked for cooldown
        msg_timestamp : float
            epoch time of message

        Returns
        -------
        bool
            command is on cooldown
        """

        # check for existing entry
        if command in self.cooldowns:
            # if less time since last call has passed than cooldown
            if msg_timestamp - self.cooldowns[command] < command.cooldown:
                return True
        # if command was not on cooldown, put on cooldown
        self.cooldowns[command] = time()
        return False



import logging

from symbot.chat.message import Message
from symbot.control.control import Control
from symbot.dev.commands._base_command import BaseCommand


class Command(BaseCommand):

    def __init__(self, control: Control):
        super().__init__(control)
        self.name = '!time'
        self.author = 'fd_symbicort'

    async def run(self, msg: Message):

        localtime = self.control.get_media('localtime')

        if localtime:

            try:
                broadcaster = self.control.environment.get('broadcaster')
            except KeyError:
                logging.info(f'{self.name} unable to find var broadcaster')
                return

            response = f'It is {await localtime.run()} for {broadcaster}'

            await self.control.respond(response)

        else:

            logging.info('Required media unavailable: localtime')



import asyncio
import logging

from symbot.chat.message import Message
from symbot.config import *


class Chat:
    """Provides an interface for Twitch chat

    Attributes
    ----------
    reader : StreamReader
        stream to download content from Twitch channel
    writer : StreamWriter
        stream to upload content to Twitch channel
    msg_queue : Queue
        thread safe queue to put downloaded content into
    resp_queue : Queue
        thread safe queue contains content to be uploaded

    Methods
    -------
    open_connection
        open an asyncio connection to a server
    read
        continuously download messages from Twitch channel
    write
        continuously upload messages to Twitch channel
    ping_pong
        handle PINGs from the server
    """

    def __init__(self):
        self.reader = None
        self.writer = None
        # thread safe queue produced here, consumed by control
        self.msg_queue = asyncio.Queue()
        # thread safe queue produced by control, consumed here
        self.resp_queue = None

    async def open_connection(self):
        """open an asyncio connection to a server"""

        # obtain reader, writer streams by connecting to server
        logging.info('connecting to server')
        self.reader, self.writer = await asyncio.open_connection(host=server, port=port)
        # log into channel with credentials
        logging.info('logging in')
        self.writer.write(f'PASS {token}\r\nNICK {nick}\r\nJOIN #{channel}\r\n'.encode('utf-8'))
        await self.writer.drain()
        # drain welcome message from reader
        logging.info('receiving welcome message')
        await self.reader.readuntil(separator='list\r\n'.encode('utf-8'))

    async def read(self):
        """continuously download messages from Twitch channel"""

        # run forever
        logging.info('awaiting messages')
        while True:
            # download message from Twitch channel
            received = (await self.reader.read(n=2048)).strip().decode('utf-8')
            # skip if PING
            if not await self._ping_pong(received):
                # parse message to workable object
                msg = Message(received)
                print(f'{msg.user}: {msg.content}')  # DEBUG
                # put to msg_queue
                # provides to control
                # MAYBE treats every message as command,
                #       because no identifying prefix is used.
                #       Monitor performance and change accordingly
                await self.msg_queue.put(msg)

    async def write(self):
        """continuously upload messages to Twitch channel"""

        # run forever
        while True:
            # wait for responses to send
            # provided by control
            resp = await self.resp_queue.get()
            # upload respond to Twitch channel
            self.writer.write(f'PRIVMSG #{channel} :{resp}\r\n'.encode('utf-8'))
            await self.writer.drain()

    async def _ping_pong(self, received):
        """handle PINGs from the server

        Parameters
        ----------
        received : str
            raw message received from Twitch channel

        Returns
        -------
        bool
            whether message was a PING
        """

        # skip if not PING
        if received.startswith('PING'):
            # respond with PONG
            self.writer.write('PONG\r\n'.encode('utf-8'))
            await self.writer.drain()
            return True
        return False
